package com.contextspace;

import java.io.File;
import java.util.Collection;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Join;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import static org.junit.Assert.*;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import com.contextspace.objectdb_example.model.Attribute;
import com.contextspace.objectdb_example.model.AttributeTypes;
import com.contextspace.objectdb_example.model.AttributeValue;
import com.contextspace.objectdb_example.model.CDI;

public class ExampleTest {
 
    private static EntityManagerFactory emf;
    private static EntityManager em;
    private static CDI createdCdi6;
    
    @BeforeClass
    public static void onceStart() throws Exception {
        emf = Persistence.createEntityManagerFactory("objectdb/db/test.odb");
        em = emf.createEntityManager();
 
        em.getTransaction().begin();
        
        CDI createdCdi1 = createCDI("baffleck", "Ben", "Geza", "Affleck");
        em.persist(createdCdi1);
        CDI createdCdi2 = createCDI("jalba", "Jessica", "Marie", "Alba");
        em.persist(createdCdi2);
        CDI createdCdi3 = createCDI("kbeckinsale", "Kate", null, "Beckinsale");
        em.persist(createdCdi3);
        CDI createdCdi4 = createCDI("abaldwin", "Alex", "Rae", "Baldwin");
        em.persist(createdCdi4);
        CDI createdCdi5 = createCDI("sbaldwin", "Stephen", "Andrew", "Baldwin");
        em.persist(createdCdi5);
        createdCdi6 = createCDI("wbaldwin", "William", null, "Baldwin");
        em.persist(createdCdi6);
        
        em.getTransaction().commit();
        
    }
    
    @AfterClass
    public static void onceEnd() throws Exception {
        
        if(em != null) {
            em.close();
        }
        
        if(emf != null) {
            emf.close();
        }
        
        File f1 = new File("objectdb/db/test.odb");
        if(f1.exists()) {
            f1.delete();
        }
    }
    
    protected static CDI createCDI(String username, String first, String middle, String last) {
        CDI cdi = new CDI();
        cdi.setCdiId(username);
        Attribute firstname = new Attribute(AttributeTypes.firstname, first);
        cdi.addAttribute(firstname);
        Attribute lastname = new Attribute(AttributeTypes.lastname, last);
        cdi.addAttribute(lastname);
        if(middle != null) {
            Attribute middlename = new Attribute(AttributeTypes.middlenames, middle);
            cdi.addAttribute(middlename);
        }
        Attribute title = new Attribute(AttributeTypes.title, "Mr");
        cdi.addAttribute(title);
      
        return cdi;
    }
    
    
    /**
     * RIGHT - this test works, but is not optimal. I've already had to do 4 joins to query 2 attributes
     * If I have to do more, its 2 more joins for each attribute. Has to be a better way.
     * 
     * @throws Exception
     */
    @Test
    public void queryAttribute1() throws Exception {
        
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<CDI> q = cb.createQuery(CDI.class);
        Root<CDI> cdi = q.from(CDI.class);
        
        Join<CDI, Attribute> attributeJoin = cdi.join("attributeList");
        Join<Attribute, AttributeValue> attributeValuesJoin = attributeJoin.join("attributeValues");
        Join<CDI, Attribute> attributeJoin2 = cdi.join("attributeList");
        Join<Attribute, AttributeValue> attributeValuesJoin2 = attributeJoin2.join("attributeValues");
        
        Predicate pred = 
            cb.and(
                cb.and(
                    cb.equal(attributeValuesJoin.get("value"), "Alex"), 
                    cb.equal(attributeJoin.get("name"), "firstname")
                ),
                cb.and(
                    cb.equal(attributeValuesJoin2.get("value"), "Baldwin"), 
                    cb.equal(attributeJoin2.get("name"), "lastname")
                )
            );
        
        q.select(cdi).where(pred);
        TypedQuery<CDI> typedQuery = em.createQuery(q);
        List<CDI> results = typedQuery.getResultList();
        System.out.println("RESULT: " + results);
        assertEquals(1, results.size());
    }
    
    /**
     * RIGHT - this test works, and its much more optimal than the first test,
     * HOWEVER, it won't work in reality. When I come to do a real search, I'm NOT
     * going to have the id of the attribute I'm searching on.
     * 
     * If I build the attribute that I pass into the cb.isMember function, it will NEVER
     * match anything. My guess being that the database is looking purely at the object
     * ids and not the content (like what java does if you don't override the equals method)
     * 
     * @throws Exception
     */
    @Test
    public void queryAttribute2() throws Exception {
        
        Attribute williamFirstNameAttribute = createdCdi6.getAttribute(AttributeTypes.firstname);
        Attribute williamLastNameAttribute = createdCdi6.getAttribute(AttributeTypes.lastname);
        
        Attribute attributeFirstName = em.getReference(Attribute.class, williamFirstNameAttribute.getId());
        Attribute attributeLastName = em.getReference(Attribute.class, williamLastNameAttribute.getId());
        
        
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<CDI> q = cb.createQuery(CDI.class);
        Root<CDI> cdi = q.from(CDI.class);
        
        /**
         * This is what I WOULD like to be able to do and pass these to the isMember function
         * but it DOES NOT work.
         */
        //Attribute attributeFirstName = new Attribute(AttributeTypes.firstname, "William");
        //Attribute attributeLastName = new Attribute(AttributeTypes.lastname, "Baldwin");
        
        Expression<Collection<Attribute>> attributes = cdi.get("attributeList");
        Predicate pred = 
            cb.and(
                cb.isMember(attributeFirstName, attributes),
                cb.isMember(attributeLastName, attributes)
            );
        
        q.select(cdi).where(pred);
        TypedQuery<CDI> typedQuery = em.createQuery(q);
        List<CDI> results = typedQuery.getResultList();
        System.out.println("RESULT: " + results);
        assertEquals(1, results.size());
    }
    
    
    /**
     * There has to be a better way of doing this????
     * @throws Exception
     */
    @Test
    public void queryAttributes3() throws Exception {
        
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<Attribute> q = cb.createQuery(Attribute.class);
        Root<Attribute> attr = q.from(Attribute.class);
        Join<Attribute, AttributeValue> attributeValueJoin = attr.join("attributeValues");
        
        Predicate pred = cb.and(
                            cb.equal(attr.get("name"), "firstname"),
                            cb.equal(attributeValueJoin.get("value"), "Alex") 
                         );
        
        q.select(attr).where(pred);
        TypedQuery<Attribute> typedQuery = em.createQuery(q);
        List<Attribute> results = typedQuery.getResultList();
        System.out.println("RESULT: " + results);
        assertEquals(1, results.size());
        
    }
}
